//#include <stm32f407xx.h>
//#include <stm32f4xx_hal_i2s.h>
//#include "codec2.h"
#include "stm32f4xx_spi.h"
#include "stm32f4xx_i2c.h"
#include "support.h"
#include "regwrites.h"
#include "pdm_filter.h"

/*
Codec2 demo for the STM32F4DISCOVERY PCB

Codec2 was developed by David Rowe (www.rowetel.com/codec2.html)

this STM32F4 implementation was written by Peter Lawrence

STMicroelectronics STM32 CPU Support package is needed in Rowley Crossworks

no warranty is provided with this code

main.c is the top-level code

dsplib.a contains the ARM CMSIS library object code
libPDMFilter_GCC.a contains the object code of a PDM to LPCM converter described in ST Micro AN3998

Codec2 source code (Rev289?) was obtained from http://freetel.svn.sourceforge.net/viewvc/freetel/codec2/
Modifications to the Codec2 code are:
fft.c was modified to use the ARM CMSIS library
FFT_ENC and FFT_DEC #defines modified in defines.h
PE_FFT_SIZE #define modified in nlp.c
GLOTTAL_FFT_SIZE #define modified in phase.c
*/

#define CODEC2_SAMPLES_PER_FRAME 80
#define CODEC2_BITS_PER_FRAME 480

#define BITS_SIZE	((CODEC2_BITS_PER_FRAME + 7) / 8)
#define CODEC2_SAMPLE_RATE 8000

#define PDM_DECIMATION_RATE 64
#define MIC_VOLUME 4
#define LPCM_DECIMATION_RATE 2
#define PDM_TO_LPCM_RATE (CODEC2_SAMPLE_RATE * LPCM_DECIMATION_RATE)
#define PDM_1MS_FRAME_SIZE  ( (PDM_TO_LPCM_RATE * PDM_DECIMATION_RATE) / 1000 / 8)

static void cs43l22_init(void);
static void mp45dt02_init(void);
static void leds_init(void);
static void toggle_gpiod(int bit);

const I2C_InitTypeDef cs43l22_i2c_params =
{
  .I2C_ClockSpeed = 100000,
  .I2C_Mode = I2C_Mode_I2C,
  .I2C_DutyCycle = I2C_DutyCycle_2,
  .I2C_OwnAddress1 = 0x33,
  .I2C_Ack = I2C_Ack_Enable,
  .I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_7bit
};

const I2S_InitTypeDef cs43l22_i2s_params =
{
  .I2S_Mode = I2S_Mode_MasterTx,
  .I2S_Standard = I2S_Standard_Phillips,
  .I2S_DataFormat = I2S_DataFormat_16b,
  .I2S_MCLKOutput = I2S_MCLKOutput_Enable,
  .I2S_AudioFreq = CODEC2_SAMPLE_RATE,
  .I2S_CPOL = I2S_CPOL_Low
};

const I2S_InitTypeDef mp45dt02_i2s_params =
{
  .I2S_Mode = I2S_Mode_MasterRx,
  .I2S_Standard = I2S_Standard_MSB,
  .I2S_DataFormat = I2S_DataFormat_16b,
  .I2S_MCLKOutput = I2S_MCLKOutput_Disable,
  .I2S_AudioFreq = PDM_TO_LPCM_RATE * PDM_DECIMATION_RATE / 32, // needed SCK rate
  .I2S_CPOL = I2S_CPOL_Low
};

/* argh... can't declare this as const because element "InternalFilter" is writable */
static PDMFilter_InitStruct mp45dt02_pdm_params =
{
  .Fs = PDM_TO_LPCM_RATE, /* rate must be double codec rate, since microphone datasheet indicates a minimum PDM rate of 1 MHz */
  .LP_HZ = 4000,
  .HP_HZ = 10,
  .In_MicChannels = 1,
  .Out_MicChannels = 1,
};

volatile static uint32_t ready_flag;
static int16_t spk_buffer[2][CODEC2_SAMPLES_PER_FRAME << 1]; /* each buffer must have double the codec samples, as it must support stereo */
static uint16_t mic_buffer[2][PDM_1MS_FRAME_SIZE >> 1];
static int16_t scratchpad[2][CODEC2_SAMPLES_PER_FRAME];

void main(void)
{
  int j;

  uint8_t bits[BITS_SIZE];
  //void *codec2;
  int16_t *dst_pnt, *scratchpad_pnt;

  /* crank the CPU up to 168 MHz */
  SetSysClock();

  /* initialize relevant peripherals */
  leds_init();
  mp45dt02_init();
  cs43l22_init();

  //codec2 = codec2_create();
  for (;;)
  {
      /* wait for ISR to signal that the scratchpad buffer is full */
      while (!ready_flag) {};
      
      scratchpad_pnt = (2 == ready_flag) ? scratchpad[1] : scratchpad[0];
      toggle_gpiod(14);
      ready_flag = 0;

      /* compression and decompression */
      if (0 == (GPIOA->IDR & 1))
      {
        GPIOD->BSRR = 1 << 12;
        //codec2_encode(codec2, bits, scratchpad_pnt);
        //codec2_decode(codec2, scratchpad_pnt, bits);
        GPIOD->BSRR = 1 << 12 << 16;

        /* pretty lights showing voicing */
        if (bits[6] & 0x80)
          GPIOD->BSRR = 1 << 13;
        else
          GPIOD->BSRR = 1 << 13 << 16;
        if (bits[6] & 0x40)
          GPIOD->BSRR = 1 << 15;
        else
          GPIOD->BSRR = 1 << 15 << 16;
      }
      else
      {
        /* turn off all Codec2 indicator LEDs */
        GPIOD->BSRR = ((1 << 12) | (1 << 13) | (1 << 15)) << 16;; 
      }

      /* copy the decoded data, expanding from mono to stereo */
      dst_pnt = (DMA1_Stream7->CR & DMA_SxCR_CT) ? spk_buffer[0] : spk_buffer[1];
      for (j = 0; j < CODEC2_SAMPLES_PER_FRAME; j++)
      {
        *dst_pnt++ = scratchpad_pnt[j];
        *dst_pnt++ = scratchpad_pnt[j];
      }

      /* enable I2S3, if not already enabled we do this here so that the transmit DMA is time-aligned to the microphone data we get */
      SPI3->I2SCFGR |= SPI_I2SCFGR_I2SE;
  }
  //codec2_destroy(codec2);

}

/*
Audio_RST : PD4
Audio_SDA : PB9
Audio_SCL : PB6
I2S3_MCK : PC7
I2S3_SCK : PC10
I2S3_SD : PC12
I2S3_WS : PA4
*/

static void cs43l22_init(void)
{
  volatile uint32_t i;

  /* Enable relevant GPIOx interface clocks */
  RCC->AHB1ENR |= RCC_AHB1ENR_GPIOAEN | RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN | RCC_AHB1ENR_GPIODEN;

  /* Configure PD4 */

  setbitpair(&GPIOD->MODER, 4, 1 /* GPO */);
  setbit32(&GPIOD->OTYPER, 4, 0 /* push-pull */);
  setbitpair(&GPIOD->OSPEEDR, 4, 2 /* 50MHz */);
  setbitpair(&GPIOD->PUPDR, 4, 0 /* no pull-up, pull-down */);

  /* Configure PB6 and PB9 */

  setbitpair(&GPIOB->MODER, 6, 2 /* AF */);
  setnibble(GPIOB->AFR, 6, 4 /* AF4 (I2C1_SCL) */);
  setbit32(&GPIOB->OTYPER, 6, 1 /* open-drain */);
  setbitpair(&GPIOB->OSPEEDR, 6, 2 /* 50MHz */);
  setbitpair(&GPIOB->PUPDR, 6, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOB->MODER, 9, 2 /* AF */);
  setnibble(GPIOB->AFR, 9, 4 /* AF4 (I2C1_SDA) */);
  setbit32(&GPIOB->OTYPER, 9, 1 /* open-drain */);
  setbitpair(&GPIOB->OSPEEDR, 9, 2 /* 50MHz */);
  setbitpair(&GPIOB->PUPDR, 9, 0 /* no pull-up, pull-down */);

  RCC->APB1ENR |= RCC_APB1ENR_I2C1EN; /* I2C1 clock enable */

  /* pulse reset of I2C1 (putting registers in a known, inactive state) */
  RCC->APB1RSTR |= RCC_APB1RSTR_I2C1RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_I2C1RST;

  /* enable I2C1 peripheral */
  I2C1->CR1 |= I2C_CR1_PE;

  I2C_Init(I2C1, (I2C_InitTypeDef *)&cs43l22_i2c_params);

  /* Configure PC7, PC10, PC12, and PA4 (I2S3 pins) */

  setbitpair(&GPIOC->MODER, 7, 2 /* AF */);
  setnibble(GPIOC->AFR, 7, 6 /* AF6 (I2S3_MCK) */);
  setbit32(&GPIOC->OTYPER, 7, 0 /* push-pull */);
  setbitpair(&GPIOC->OSPEEDR, 7, 2 /* 50MHz */);
  setbitpair(&GPIOC->PUPDR, 7, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOC->MODER, 10, 2 /* AF */);
  setnibble(GPIOC->AFR, 10, 6 /* AF6 (I2S3_CK) */);
  setbit32(&GPIOC->OTYPER, 10, 0 /* push-pull */);
  setbitpair(&GPIOC->OSPEEDR, 10, 2 /* 50MHz */);
  setbitpair(&GPIOC->PUPDR, 10, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOC->MODER, 12, 2 /* AF */);
  setnibble(GPIOC->AFR, 12, 6 /* AF6 (I2S3_SD) */);
  setbit32(&GPIOC->OTYPER, 12, 0 /* push-pull */);
  setbitpair(&GPIOC->OSPEEDR, 12, 2 /* 50MHz */);
  setbitpair(&GPIOC->PUPDR, 12, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOA->MODER, 4, 2 /* AF */);
  setnibble(GPIOA->AFR, 4, 6 /* AF6 (I2S3_WS) */);
  setbit32(&GPIOA->OTYPER, 4, 0 /* push-pull */);
  setbitpair(&GPIOA->OSPEEDR, 4, 2 /* 50MHz */);
  setbitpair(&GPIOA->PUPDR, 4, 0 /* no pull-up, pull-down */);

  /* pulse codec reset line (active low) */
  GPIOD->BSRR = 1 << 4 << 16;
  for (i = 0; i < 32000; i++) {};
  GPIOD->BSRR = 1 << 4;

  i = Codec_ReadRegister(0x01);
  Codec_WriteRegister(0x02, 0x01); /* Keep Codec powered OFF */
  Codec_WriteRegister(0x04, 0xAF); /* SPK always OFF & HP always ON */
  Codec_WriteRegister(0x05, 0x81); /* Clock configuration: Auto detection */
  Codec_WriteRegister(0x06, 0x04); /* configure for I2S interface */  
  Codec_WriteRegister(0x20, 0x00);
  Codec_WriteRegister(0x21, 0x00);
  Codec_WriteRegister(0x02, 0x9E); /* Power on the Codec */
  Codec_WriteRegister(0x27, 0x00); /* Disable the limiter attack level */
  Codec_WriteRegister(0x1F, 0x0F); /* Adjust Bass and Treble levels */
  Codec_WriteRegister(0x1A, 0x0A); /* Adjust PCM volume level */
  Codec_WriteRegister(0x1B, 0x0A);

#if 0 // digital mic test
  /* Enable pass-through on AIN4A and AIN4B */
  Codec_WriteRegister(0x08, 0x08);
  Codec_WriteRegister(0x09, 0x08);
  
  /* Route the analog input to the HP line */
  Codec_WriteRegister(0x0E, 0xC0);
  
  /* Set the pass-though volume */
  Codec_WriteRegister(0x14, 0x00);
  Codec_WriteRegister(0x15, 0x00);
#endif

  /* Enable clock to SPI3/I2S3 */
  RCC->APB1ENR |= RCC_APB1ENR_SPI3EN;

  /* pulse SPI3 reset (putting registers in a known, inactive state) */
  RCC->APB1RSTR |= RCC_APB1RSTR_SPI3RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_SPI3RST;

  HAL_I2S_Init(SPI3, (I2S_InitTypeDef *)&cs43l22_i2s_params);

  /* enable DMA1 clock */
  RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN;

  /* disable DMA1 Stream7 before configuring */
  do
  {
    DMA1_Stream7->CR &= ~DMA_SxCR_EN;
  } while (DMA1_Stream7->CR & DMA_SxCR_EN);

  /* configure DMA1 Stream7 (following order prescribed in section 8.3.17 of RM) */

  DMA1_Stream7->PAR = (uint32_t)&(SPI3->DR);
  DMA1_Stream7->M0AR = (uint32_t)spk_buffer[0];
  DMA1_Stream7->M1AR = (uint32_t)spk_buffer[1];
  DMA1_Stream7->NDTR = CODEC2_SAMPLES_PER_FRAME << 1;

  DMA1_Stream7->CR = (0 /* channel 0 */ << 25);

  DMA1_Stream7->FCR = DMA_SxFCR_DMDIS | DMA_SxFCR_FTH;

  DMA1_Stream7->CR = 
    (0 /* channel 0 */ << 25) | 
    DMA_SxCR_DBM | 
    (1 << 16) | /* priority */
    (1 /* 16-bit memory */ << 13) | 
    (1 /* 16-bit peripheral */ << 11) | 
    DMA_SxCR_MINC | 
    (1 /* memory-to-peripheral */ << 6) |
    DMA_SxCR_TCIE |
    DMA_SxCR_EN;

  SPI3->CR2 |= SPI_CR2_TXDMAEN;

  NVIC_SetPriority(DMA1_Stream7_IRQn, 1);
  NVIC_EnableIRQ(DMA1_Stream7_IRQn);

#if 0
  /* enable I2S3 */
  SPI3->I2SCFGR |= SPI_I2SCFGR_I2SE;
#endif

}

void DMA1_Stream7_IRQHandler(void)
{
  ready_flag = (DMA1_Stream7->CR & DMA_SxCR_CT) ? 1 : 2;

  /* acknowledge interrupt */
  DMA1->HIFCR = DMA_HIFCR_CTCIF7;
}

static void mp45dt02_init(void)
{
  /* Enable relevant GPIOx interface clocks */
  RCC->AHB1ENR |= RCC_AHB1ENR_GPIOBEN | RCC_AHB1ENR_GPIOCEN;

  /* Configure PC3 and PB10 (I2S2 pins) */

  setbitpair(&GPIOC->MODER, 3, 2 /* AF */);
  setnibble(GPIOC->AFR, 3, 5 /* AF5 (I2S2_SD) */);
  setbit32(&GPIOC->OTYPER, 3, 0 /* push-pull */);
  setbitpair(&GPIOC->OSPEEDR, 3, 2 /* 50MHz */);
  setbitpair(&GPIOC->PUPDR, 3, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOB->MODER, 10, 2 /* AF */);
  setnibble(GPIOB->AFR, 10, 5 /* AF5 (I2S2_CK) */);
  setbit32(&GPIOB->OTYPER, 10, 0 /* push-pull */);
  setbitpair(&GPIOB->OSPEEDR, 10, 2 /* 50MHz */);
  setbitpair(&GPIOB->PUPDR, 10, 0 /* no pull-up, pull-down */);

  /* Enable clock to CRC (prerequisite to calling PDM library) */
  RCC->AHB1ENR |= RCC_AHB1LPENR_CRCLPEN;

  PDM_Filter_Init(&mp45dt02_pdm_params);

  /* Enable clock to SPI2/I2S2 */
  RCC->APB1ENR |= RCC_APB1ENR_SPI2EN;

  /* pulse SPI2 reset (putting registers in a known, inactive state) */
  RCC->APB1RSTR |= RCC_APB1RSTR_SPI2RST;
  RCC->APB1RSTR &= ~RCC_APB1RSTR_SPI2RST;

  HAL_I2S_Init(SPI2, (I2S_InitTypeDef *)&mp45dt02_i2s_params);

  /* enable DMA1 clock */
  RCC->AHB1ENR |= RCC_AHB1ENR_DMA1EN;

  /* disable DMA1 Stream3 before configuring */
  do
  {
    DMA1_Stream3->CR &= ~DMA_SxCR_EN;
  } while (DMA1_Stream3->CR & DMA_SxCR_EN);

  /* configure DMA1 Stream3 (following order prescribed in section 8.3.17 of RM) */

  DMA1_Stream3->PAR = (uint32_t)&(SPI2->DR);
  DMA1_Stream3->M0AR = (uint32_t)mic_buffer[0];
  DMA1_Stream3->M1AR = (uint32_t)mic_buffer[1];
  DMA1_Stream3->NDTR = sizeof(mic_buffer) / 4;

  DMA1_Stream3->CR = (0 /* channel 0 */ << 25);

  DMA1_Stream3->FCR = DMA_SxFCR_DMDIS | DMA_SxFCR_FTH;

  DMA1_Stream3->CR = 
    (0 /* channel 0 */ << 25) | 
    DMA_SxCR_DBM | 
    (0 << 16) | /* priority */
    (1 /* 16-bit memory */ << 13) | 
    (1 /* 16-bit peripheral */ << 11) | 
    DMA_SxCR_MINC | 
    (0 /* peripheral-to-memory */ << 6) |
    DMA_SxCR_TCIE |
    DMA_SxCR_EN;

  SPI2->CR2 |= SPI_CR2_RXDMAEN;

  NVIC_SetPriority(DMA1_Stream3_IRQn, 1);
  NVIC_EnableIRQ(DMA1_Stream3_IRQn);

  /* enable I2S2 */
  SPI2->I2SCFGR |= SPI_I2SCFGR_I2SE;
}

static void leds_init(void)
{
  /* Enable relevant GPIOx interface clocks */
  RCC->AHB1ENR |= RCC_AHB1ENR_GPIODEN;

  /* Configure PD12, PD13, PD14, PD15 (LED pins) */

  setbitpair(&GPIOD->MODER, 12, 1 /* GPO */);
  setbit32(&GPIOD->OTYPER, 12, 0 /* push-pull */);
  setbitpair(&GPIOD->OSPEEDR, 12, 2 /* 50MHz */);
  setbitpair(&GPIOD->PUPDR, 12, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOD->MODER, 13, 1 /* GPO */);
  setbit32(&GPIOD->OTYPER, 13, 0 /* push-pull */);
  setbitpair(&GPIOD->OSPEEDR, 13, 2 /* 50MHz */);
  setbitpair(&GPIOD->PUPDR, 13, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOD->MODER, 14, 1 /* GPO */);
  setbit32(&GPIOD->OTYPER, 14, 0 /* push-pull */);
  setbitpair(&GPIOD->OSPEEDR, 14, 2 /* 50MHz */);
  setbitpair(&GPIOD->PUPDR, 14, 0 /* no pull-up, pull-down */);

  setbitpair(&GPIOD->MODER, 15, 1 /* GPO */);
  setbit32(&GPIOD->OTYPER, 15, 0 /* push-pull */);
  setbitpair(&GPIOD->OSPEEDR, 15, 2 /* 50MHz */);
  setbitpair(&GPIOD->PUPDR, 15, 0 /* no pull-up, pull-down */);
}

void DMA1_Stream3_IRQHandler(void)
{
  static uint16_t *src_pnt;
  static int16_t pdm_temp[PDM_TO_LPCM_RATE / 1000];
  static int scratchpad_index, j, scratchpad_bank;

  src_pnt = (DMA1_Stream3->CR & DMA_SxCR_CT) ? mic_buffer[0] : mic_buffer[1];

#if (64 == PDM_DECIMATION_RATE)
  PDM_Filter_64_MSB
#elif (80 == PDM_DECIMATION_RATE)
  PDM_Filter_80_MSB
#else
  #error bad PDM_DECIMATION_RATE
#endif
  ((uint8_t *) src_pnt, pdm_temp, MIC_VOLUME, &mp45dt02_pdm_params);

  /* audio already low-pass filtered; down-sample for Codec code */
  for (j = 0; j < (sizeof(pdm_temp) / sizeof(uint16_t)); j += LPCM_DECIMATION_RATE)
  {
    scratchpad[scratchpad_bank][scratchpad_index++] = pdm_temp[j];
    if (scratchpad_index >= CODEC2_SAMPLES_PER_FRAME)
    {
      scratchpad_index = 0;
      /* signal Codec code */
      ready_flag = (scratchpad_bank) ? 2 : 1;
      scratchpad_bank = (scratchpad_bank) ? 0 : 1;
      break;
    }
  }

  /* acknowledge interrupt */
  DMA1->LIFCR = DMA_LIFCR_CTCIF3;
}

static void toggle_gpiod(int bit)
{
  if (GPIOD->ODR & (1 << bit))
    GPIOD->BSRR = 1 << bit;
  else
    GPIOD->BSRR = 1 << bit;
}
